////////////
//
// ParserCPP.cpp
//
// In this one, I will not be usign state style parsing
/////////////////


#include "StdAfx.h"

#include "ParserCPP.h"

#include <locale>

#include <assert.h>

//
// PUBLICS
//


const char ParserCPP::_sOperators[] = "!%&()*+,-./:;<=>?[]^{|}~";
const int ParserCPP::_sNumOperators = sizeof (ParserCPP::_sOperators) /
											sizeof (ParserCPP::_sOperators[0]);




ParserCPP::ParserCPP (std::ifstream& inFile, std::ofstream& outFile) 
		: _inFile (inFile),
		  _outFile (outFile),
		  _multiLineCommentStarted (false) ,
		  _isDoubleQuoteString (false),
		  _parserState (STATE_START),
		  _len (0)
{
	
	// read the keywords file and store the keywords in the set
	std::ifstream keywordsFile ("C_Keywords.txt");
	while (! keywordsFile.eof())
	{
		std::string kword;
		std::getline (keywordsFile, kword);
		if (!kword.empty())
			_keywords.insert (kword);
	}
	keywordsFile.close();
}
 
// -------------------------------------------------------------------------
// PARSING NOTES/TIPS:
//
//
// for multiple charachter runs, note down start end indices
// for single charachter runs write them out right away

//
// NOTES:
// Parsing must be done on rules and rules have precedence
// in general multiple combinatation rules have presidence over single charachter words
// so we will start by checking them first

// in start state we can go either way
// so we need to start checking in precendece order of rules
// implying that we first must check compund rules
// before checking simpler ones

//
// APPROACH:
// Read off a line, and then analyze it charachter by charachter
// -------------------------------------------------------------------------



// -------------------------------------------------------------------------------------------------------
// NOTES:
// 
// Switch Board state advantages
// All initial transition rules are in one place, rather than being scattered all over the place
//
// this is outside line reading because states can span multiple lines
// e.g multi line comment, and line continuation charachter
//
// In switchBoard rules precedence must be taken into account ...
// if same letter can appear as the first letter of more than two different rules
// then the compound rules must be checked first!
//
// -------------------------------------------------------------------------------------------------------

void
ParserCPP::parse()
{
	_parserState = STATE_START;
	

			
	while (!_inFile.eof())		
	{
		// read upto size-1 or new line but do not extract the new line
		// the string read is null terminated
		_inFile.getline (_line, sizeof(_line) - 3, '\n');	//-3 to sleave some space for sentinal in the buffer
		if (_inFile.gcount() == 0)
			continue;		// some error on the line therefore continue t read the next line

		_len = strlen (_line);
		_line[_len+1] = '\0';	// sentinal values .. they make the code for detecting sequence of charachters 
		_line[_len+2] = '\0';	// at the end of line easier to detect .. becasue we know some values in the 
		_line[_len+3] = '\0';	// buffer exist beyond the _len field -- 
		int idx = 0;
		while (idx < _len)
		{
            switch (_parserState)
			{
				case STATE_START:		// -- the SwitchBoard state --
				{
					char c = _line [idx];

					if ( isCommentsStart(idx))
						_parserState = STATE_Comment;

					else if (isStringSequence (idx))
						_parserState = STATE_String;

					else if (c == '#')
						_parserState = STATE_Preprocessor;

					else if (isOperator (c))
						_parserState = STATE_Operator;

					else if (isNumber(c))
						_parserState = STATE_Number;

					else if (isspace (c))
						_parserState = STATE_WhiteSpace;

					else if (isalpha(c) || c == '_')
						_parserState = STATE_IdentifierOrKeyword;

					else
					{
                        assert (false);		
						// no other case should be possible
						// but since I have not taken into account all possible scenarios
						// we must still be prepared to handle if we do come in here!
						// Therefore write out this single charachter as we
						// have failed to recognize it in any of the rules
						_outFile << _line[idx];
						++idx;
					}
				}
				break;


				case STATE_Operator:
				{
					int j = idx;			// record the start position;
                    while (idx < _len)		// either do idx < len or do _iine[idx] == NULL -- same thing
					{
						char c = _line[idx];
						if ( isOperator (c) )
                            idx++;
						else
							break;
					}
					writeText (j, idx-1, SPAN_TYPE_Operator);
					_parserState = STATE_START;
				}
				break;


				case STATE_Number:
				{
					int j = idx;			// record the start position;
                    while (idx < _len)		// either do idx < len or do _iine[idx] == NULL -- same thing
					{
						char c = _line[idx];
						if (isdigit(c) || c =='x' || c == 'o' || c == '.' || 
										tolower(c) == 'f' || tolower(c) == 'e' || c == '+' || c == '-')
                            idx++;
						else
							break;
					}
					writeText (j, idx-1, SPAN_TYPE_Number);
					_parserState = STATE_START;
				}
				break;

				
				case STATE_Preprocessor:
				{
					// end of a preprocessor is marked by a space
					int j = idx;			// record the start position;
                    while (idx < _len && !isspace (_line[idx]))		
						idx++;

					writeText (j, idx-1, SPAN_TYPE_Preprocessor);
					_parserState = STATE_START;
				}
				break;

				
				case STATE_WhiteSpace:
				{
					while (idx < _len)
					{
						if (isspace (_line[idx]))		
						{
							if (_line[idx] == '\t')
                                _outFile << "    ";		// change tabs to 4 spaces!
							else
								_outFile << (_line[idx]);

							++idx;
						}
						else
							break;
					}
					_parserState = STATE_START;
				}
				break;

						
				case STATE_IdentifierOrKeyword:
				{
					// a space or an operator stops the identifier
					int j = idx;			// record the start position;
                    while ( idx < _len && !isspace (_line[idx]) && ! isOperator (_line[idx]) )		
						++idx;

					// determine if the word is one of the keywords or not
					//
					char copied = _line[idx];
					_line[idx] = '\0';	
					if (isKeyword (_line + j))
						writeText (j, idx-1, SPAN_TYPE_Keyword);
					else
						writeText (j, idx-1, SPAN_TYPE_Identifier);
                    
					// restore charachter in line's  buffer!
					_line[idx] = copied;		// restore character

					_parserState = STATE_START;
				}
				break;

				
				case STATE_String:
				{
					// first note down what string type is
					// it is important to detect other type intermixed 
					// inbetwen double quotes
					if (_line[idx] == '\"')
						_isDoubleQuoteString = true;
					
					int j = idx;		// starting " or '
					idx++;
					while (idx < _len)
					{
						if (isStringSequence (idx))	 
						{
							if ( ! (_isDoubleQuoteString && _line[idx] == '\'' ))
							{
								_isDoubleQuoteString = false;
								++idx;	// so that we can use the same code (idx-1) in writeText() call
								break;
							}
						}
						++idx;
					}

					if (idx == _len)
					{
						// and we have not found closing ", it means it is a line continuation case
						// and it should be the last non-white charachter on the line
						// therefore search in reverse and find the continuation charachter's backslash index
						// on the line (there could be white spaces, that is why search is required)
                        int idx_line_cont_char = findIndexOfLineContinuationCharachter ();
						if (idx_line_cont_char == -1)	
						{
							// no line continuation charachter
							// which is syntax error in C++ code anywas, therefore treat how so ever!
							writeText (j, idx-1, SPAN_TYPE_String);
							_parserState = STATE_START;
						}
						else
						{
							writeText (j, idx_line_cont_char-1, SPAN_TYPE_String);
							_outFile << '\\';
							// stirng stat would continue therefore no change in parser state
						}
					}
					
					else
					{
						writeText (j, idx-1, SPAN_TYPE_String);
						_parserState = STATE_START;
					}
				}
				break;


				case STATE_Comment:
					handleComments (idx);		// note - passed by ref!
				break;

				
				default:
					assert (false);
					idx++;
				break;

			}		// end switch
		}	// end while (line)
		
		_outFile << "\n";	
	}	// end while (file)
}





bool
ParserCPP::isKeyword(const char* word) const
{
	std::string str (word);
	Keywords::const_iterator it = _keywords.find (str);
	if (it != _keywords.end())
		return (true);
	else
		return (false);
}



bool
ParserCPP::isOperator (const char c) const
{
	for (int i=0; i<_sNumOperators; i++)
		if (_sOperators[i] == c)
			return (true);

	return (false);
}



bool
ParserCPP::isNumber (const char c) const
{
	if ( (c >= '0' && c <= '9') || c == '.') 
		return (true);
	else
		return (false);
}



bool
ParserCPP::isStringSequence (int idx) const
{
	// treat single quoted (charachters i.e.) as strings as well!

	char c = _line [idx];
	if (c == '\"' || c == '\'')		// equal to "  or '
	{
		if (idx == 0)		// definite string start
			return (true);
		else	// ( i.e. when idx > 0)
		{
			// check the letter previous to it to see if it is not a backlash!
			char c2 = _line [idx-1];
			if (c2 == '\\')
				return (false);
			else
				return (true);
		}
	}
	else
		return (false);
}

int
ParserCPP::findIndexOfLineContinuationCharachter()
{
	int k = _len - 1;
	while (isspace( _line[k]))		
		k--;

	if (_line[k] != '\\')
		return (-1);
	else
		return (k);
}

bool
ParserCPP::isCommentsStart (int idx)  const
{
	if (_line[idx] == '/' && (_line[idx+1] == '/' || _line[idx+1] == '*'))
		return (true);
	else
		return (false);
}



void
ParserCPP::handleComments (int& idx)
{
	int j = idx;

	if (_multiLineCommentStarted)
	{
		// scan the line for end sequence, and if not found ,write all the text
		bool end_sequence_found = false;
		while (idx < _len)
		{
            if (_line[idx] == '*' && _line[idx+1] =='/')
			{
				idx += 2;
				_multiLineCommentStarted = false;
				_parserState = STATE_START;
				break;
			}
			++idx;
		}
		writeText (j, idx-1, SPAN_TYPE_Comment);
	}

	else
	{
        if (_line[idx+1] == '/')		// single line comment
		{
			idx = _len;
			writeText (j, idx-1, SPAN_TYPE_Comment);
			_parserState = STATE_START;
			return;
		}

		else		// 2nd charachter must be *
		{
			_multiLineCommentStarted = true;
			handleComments (idx);		// recurse!
		}
	}
}



// -------------------------------------------------------------------------------------
//

void
ParserCPP::writeText (int idx1, int idx2)
{
	// idx1 and idx2 are indices at the first and last letter to be written 
	char previous = _line[idx2+1];
	_line[idx2+1] = '\0';		// thus making it a null terminating charachter
	_outFile <<  & _line[idx1];
	_line[idx2+1] = previous;
}



void
ParserCPP::writeText (int idx1, int idx2, SpanType st)
{
	// st is the class enum for the class name of the span tag
	if (st != SPAN_TYPE_NONE)
		_outFile << "<span class=\"" << SpanClassName [st] << "\">";

    writeText (idx1, idx2);
	
	if (st != SPAN_TYPE_NONE)
		_outFile <<"</span>";
}






