//////////////
//
// ParserCPP.h
//
/////////////////////



#pragma once

#include <map>
#include <list>
#include <set>
#include <string>


// enums to identify the different span tag classes
typedef enum
{
	SPAN_TYPE_NONE = -1,

	SPAN_TYPE_Identifier = 0,		// for everyhting else (variable,s spaces, anything)
	SPAN_TYPE_Comment,				// or /* */ style comments
	SPAN_TYPE_Keyword,
	SPAN_TYPE_Operator,
	SPAN_TYPE_Number,
	SPAN_TYPE_String,
	SPAN_TYPE_Preprocessor,			// for # ... things!

	NUM_SPAN_TYPES,

	// the following states can be used as state types for the 
} SpanType;

// and their string equivalants
const char SpanClassName[][64] = {	"Identifier",
									"Comment",
									"Keyword",
									"Operator",
									"Number",
									"String",
									"Preprocessor"
								};




// ----------------- PARSER STAT ENUMS ---------------
// these are the state parser state machine can be in at any one time
typedef enum
{
	STATE_START = -1,

	STATE_WhiteSpace = 0,
	STATE_IdentifierOrKeyword,
	STATE_Operator,
	STATE_String,
	STATE_Number,
	STATE_Preprocessor,
	STATE_Comment,

	NUM_STATES

} ParserState;
	



// -------------------------------------------------------------------------

class ParserCPP
{
	typedef std::set<std::string> Keywords;
	Keywords _keywords;

	static const char _sOperators[];
	static const int  _sNumOperators;

	// enum to define what types
	// this also serves as the states style parsing as well
	
	char _line[2048];
	int _len;
	ParserState _parserState;	// current parsing state
	

	std::ifstream& _inFile;
	std::ofstream& _outFile;

	// comments are tricky so need to maintain some state between reading lines
	bool _multiLineCommentStarted;
	bool _isDoubleQuoteString;




	bool isKeyword(const char* word) const;
	bool isOperator (const char c) const;
	bool isNumber (const char c) const;

	bool isStringSequence (int idx) const;
	int findIndexOfLineContinuationCharachter();

	bool isCommentsStart (int idx) const;
	void handleComments (int& idx);		// note: pas by ref!


	void writeText (int idx1, int idx2);
	void writeText (int idx1, int idx2, SpanType st);
	// write out the charachters from idx1 to id2 inclusive 
	// one uses no span tag, other uses the specified span tag
	

public:
	explicit ParserCPP (std::ifstream& inFile, std::ofstream& outFile);

	void parse();		
	// older (now commented) routine used the push model ... where character was read
	// and then handed over to the states ... As I have noted in my notes, it was not 
	// very good for this ... For parsing, pull model works best, where states pull whatever
	// and however much data they need!


};


